#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from pythonwrench.collections import dump_dict
from pythonwrench.typing import BuiltinScalar, SupportsIterLen
from torch import Tensor, nn

from torchwrench.core.make import DeviceLike, DTypeLike, GeneratorLike
from torchwrench.extras.numpy import np
from torchwrench.nn.functional.transform import (
    PadCropAlign,
    PadMode,
    PadValue,
    SqueezeMode,
    T_BuiltinScalar,
    as_tensor,
    flatten,
    identity,
    move_to_rec,
    pad_and_crop_dim,
    repeat_interleave_nd,
    resample_nearest_freqs,
    resample_nearest_rates,
    resample_nearest_steps,
    shuffled,
    squeeze,
    to_item,
    top_p,
    topk,
    transform_drop,
    unsqueeze,
    view_as_complex,
    view_as_real,
)
from torchwrench.types._typing import (
    ComplexFloatingTensor,
    LongTensor,
    ScalarLike,
    T_TensorOrArray,
)
from torchwrench.utils import return_types

from .module import EModule, Module

T = TypeVar("T")


class AsTensor(Module):
    """
    Module version of :func:`~to.as_tensor`.
    """

    def __init__(
        self,
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = None,
    ) -> None:
        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: Any) -> Tensor:
        return as_tensor(x, dtype=self.dtype, device=self.device)

    def extra_repr(self) -> str:
        return dump_dict(
            dtype=self.dtype,
            device=self.device,
            ignore_lst=(None,),
        )


class Flatten(Module):
    def __init__(self, start_dim: int = 0, end_dim: Optional[int] = None) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.flatten`.
        """
        super().__init__()
        self.start_dim = start_dim
        self.end_dim = end_dim

    @overload
    def forward(self, x: Tensor) -> Tensor:  # type: ignore
        ...

    @overload
    def forward(self, x: Union[np.ndarray, np.generic]) -> np.ndarray: ...

    @overload
    def forward(self, x: T_BuiltinScalar) -> List[T_BuiltinScalar]: ...

    @overload
    def forward(self, x: Iterable[T_BuiltinScalar]) -> List[T_BuiltinScalar]:  # type: ignore
        ...

    def forward(self, x: Any) -> Any:
        return flatten(
            x,
            start_dim=self.start_dim,
            end_dim=self.end_dim,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            start_dim=self.start_dim,
            end_dim=self.end_dim,
        )


class Identity(Module):
    def __init__(self, *args, **kwargs) -> None:
        """Identity class placeholder.

        Unlike torch.nn.Identity which only supports Tensor typing, its type output is the same than its input type.
        """
        super().__init__()

    def forward(self, x: T) -> T:
        return identity(x)


class MoveToRec(Module):
    """
    Module version of :func:`~torchwrench.move_to_rec`.
    """

    def __init__(
        self,
        predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    ) -> None:
        super().__init__()
        self.predicate = predicate

    def forward(self, x: Any) -> Any:
        return move_to_rec(x, predicate=self.predicate)


class PadAndCropDim(Module):
    def __init__(
        self,
        target_length: int,
        align: PadCropAlign = "left",
        pad_value: PadValue = 0.0,
        dim: int = -1,
        mode: PadMode = "constant",
        generator: GeneratorLike = None,
    ) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.pad_and_crop_dim`.
        """
        super().__init__()
        self.target_length = target_length
        self.align: PadCropAlign = align
        self.pad_value = pad_value
        self.dim = dim
        self.mode: PadMode = mode
        self.generator: GeneratorLike = generator

    def forward(self, x: Tensor) -> Tensor:
        return pad_and_crop_dim(
            x,
            self.target_length,
            align=self.align,
            pad_value=self.pad_value,
            dim=self.dim,
            mode=self.mode,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            target_length=self.target_length,
            align=self.align,
            pad_value=self.pad_value,
            dim=self.dim,
            mode=self.mode,
        )


class RepeatInterleaveNd(Module):
    """
    For more information, see :func:`~to.repeat_interleave_nd`.
    """

    def __init__(self, repeats: int, dim: int) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim

    def forward(self, x: Tensor) -> Tensor:
        return repeat_interleave_nd(x, self.repeats, self.dim)

    def extra_repr(self) -> str:
        return dump_dict(repeats=self.repeats, dim=self.dim)


class ResampleNearestRates(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.transform.resample_nearest_rates`.
    """

    def __init__(
        self,
        rates: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        super().__init__()
        self.rates = rates
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_rates(
            x,
            rates=self.rates,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(rates=self.rates, dims=self.dims)


class ResampleNearestFreqs(Module):
    def __init__(
        self,
        orig_freq: int,
        new_freq: int,
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.resample_nearest_freqs`.
        """
        super().__init__()
        self.orig_freq = orig_freq
        self.new_freq = new_freq
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_freqs(
            x,
            orig_freq=self.orig_freq,
            new_freq=self.new_freq,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            orig_freq=self.orig_freq, new_freq=self.new_freq, dims=self.dims
        )


class ResampleNearestSteps(Module):
    def __init__(
        self,
        steps: Union[float, Iterable[float]],
        dims: Union[int, Iterable[int]] = -1,
        round_fn: Callable[[Tensor], Tensor] = torch.floor,
    ) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.resample_nearest_steps`.
        """
        super().__init__()
        self.steps = steps
        self.dims = dims
        self.round_fn = round_fn

    def forward(self, x: Tensor) -> Tensor:
        return resample_nearest_steps(
            x,
            steps=self.steps,
            dims=self.dims,
            round_fn=self.round_fn,
        )

    def extra_repr(self) -> str:
        return dump_dict(steps=self.steps, dims=self.dims)


class Squeeze(Module):
    """
    Module version of :func:`~torchwrench.squeeze`.
    """

    def __init__(
        self,
        dim: Union[int, Iterable[int], None] = None,
        mode: SqueezeMode = "view_if_possible",
    ) -> None:
        super().__init__()
        self.dim = dim
        self.mode: SqueezeMode = mode

    def forward(self, x: Tensor) -> Tensor:
        return squeeze(x, self.dim, self.mode)

    def extra_repr(self) -> str:
        return dump_dict(
            dim=self.dim,
            mode=self.mode,
        )


class Shuffled(Module):
    def __init__(
        self,
        dims: Union[int, Iterable[int]] = -1,
        generator: GeneratorLike = None,
    ) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.shuffled`.
        """
        super().__init__()
        self.dims = dims
        self.generator: GeneratorLike = generator

    def forward(self, x: Tensor) -> Tensor:
        return shuffled(x, dims=self.dims, generator=self.generator)

    def extra_repr(self) -> str:
        return dump_dict(dims=self.dims)


class ToItem(Module):
    """
    Module version of :func:`~torchwrench.to_item`.
    """

    def forward(
        self,
        x: Union[ScalarLike, Tensor, np.ndarray, SupportsIterLen],
    ) -> BuiltinScalar:
        return to_item(x)  # type: ignore


class TransformDrop(Generic[T], EModule[T, T]):
    def __init__(
        self,
        transform: Callable[[T], T],
        p: float,
        generator: GeneratorLike = None,
    ) -> None:
        """
        For more information, see :func:`~torchwrench.nn.functional.transform.transform_drop`.
        """
        super().__init__()
        self.transform = transform
        self.p = p
        self.generator: GeneratorLike = generator

    def forward(self, x: T) -> T:
        return transform_drop(
            transform=self.transform,
            x=x,
            p=self.p,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(p=self.p)


class Unsqueeze(Module):
    """
    Module version of :func:`~torchwrench.unsqueeze`.
    """

    def __init__(
        self,
        dim: Union[int, Iterable[int]],
        mode: SqueezeMode = "view_if_possible",
    ) -> None:
        super().__init__()
        self.dim = dim
        self.mode: SqueezeMode = mode

    def forward(self, x: T_TensorOrArray) -> T_TensorOrArray:
        return unsqueeze(x, self.dim, self.mode)

    def extra_repr(self) -> str:
        return dump_dict(dim=self.dim, mode=self.mode)


class ViewAsReal(Module):
    """
    Module version of :func:`~torchwrench.to_item`.
    """

    def forward(
        self, x: Union[Tensor, np.ndarray, complex]
    ) -> Union[Tensor, np.ndarray, Tuple[float, float]]:
        return view_as_real(x)


class ViewAsComplex(Module):
    """
    Module version of :func:`~torchwrench.to_item`.
    """

    def forward(
        self, x: Union[Tensor, np.ndarray, Tuple[float, float]]
    ) -> Union[ComplexFloatingTensor, np.ndarray, complex]:
        return view_as_complex(x)


class Topk(Module):
    """
    Module version of :func:`~torchwrench.topk`.
    """

    def __init__(
        self,
        k: int,
        dim: int = -1,
        largest: bool = True,
        sorted: bool = True,
        *,
        return_values: bool = True,
        return_indices: bool = True,
    ) -> None:
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

        super().__init__()
        self.k = k
        self.dim = dim
        self.largest = largest
        self.sorted = sorted
        self.return_values = return_values
        self.return_indices = return_indices

    def forward(self, x: Tensor) -> Union[Tensor, LongTensor, return_types.topk]:
        return topk(
            x=x,
            k=self.k,
            dim=self.dim,
            largest=self.largest,
            sorted=self.sorted,
            return_values=self.return_values,
            return_indices=self.return_indices,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            k=self.k,
            dim=self.dim,
            largest=self.largest,
            sorted=self.sorted,
            return_values=self.return_values,
            return_indices=self.return_indices,
        )


class TopP(Module):
    """
    Module version of :func:`~torchwrench.top_p`.
    """

    def __init__(
        self,
        p: float,
        dim: int = -1,
        largest: bool = True,
        *,
        return_values: bool = True,
        return_indices: bool = True,
    ) -> None:
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

        super().__init__()
        self.p = p
        self.dim = dim
        self.largest = largest
        self.return_values = return_values
        self.return_indices = return_indices

    def forward(self, x: Tensor) -> Union[Tensor, LongTensor, return_types.top_p]:
        return top_p(
            x=x,
            p=self.p,
            dim=self.dim,
            largest=self.largest,
            return_values=self.return_values,
            return_indices=self.return_indices,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            p=self.p,
            dim=self.dim,
            largest=self.largest,
            return_values=self.return_values,
            return_indices=self.return_indices,
        )


# Aliases
ToTensor = AsTensor
TopK = Topk



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor

from torchwrench.core.make import DeviceLike, as_device
from torchwrench.nn import functional as F
from torchwrench.nn.functional.padding import pad_and_stack_rec, pad_dim
from torchwrench.types import BoolTensor, LongTensor


def activity_to_segments(x: Tensor) -> LongTensor:
    """Extracts segments start and end positions from a boolean activity/mask tensor.

    Example 1
    ----------
    >>> x = torch.as_tensor([0, 1, 1, 0, 0, 1, 1, 1, 1, 0]).bool()
    >>> starts, ends = extract_segments(x)
    >>> starts
    ... tensor([1, 5])
    >>> ends
    ... tensor([3, 9])

    Example 2
    ----------
    >>> x = torch.as_tensor([[1, 1, 1, 0], [1, 0, 0, 1]]).bool()
    >>> indices, starts, ends = extract_segments(x)
    >>> indices
    ... tensor([0, 1, 1])
    >>> starts
    ... tensor([0, 0, 3])
    >>> ends
    ... tensor([3, 1, 4])

    Args:
        x: (..., N) bool tensor containing D dims

    Returns:
        segments: (D+1, M) tensor, where M is the total number of segments
            When D > 1, segments also contains indices of the source column for each start and end value. See Example 2 for details.
    """
    if not isinstance(x, BoolTensor):
        msg = f"Invalid argument {x=}. (expected BoolTensor)"
        raise ValueError(msg)

    x = x.int()
    x = pad_dim(x, x.shape[-1] + 2, align="center", pad_value=0, dim=-1)
    diff = x[..., 1:] - x[..., :-1]

    starts = torch.stack(torch.where(diff > 0))
    ends = torch.stack(torch.where(diff < 0))

    result = torch.cat([starts, ends[-1:]], dim=0)
    return result  # type: ignore


def segments_to_segments_list(
    segments: Tensor,
    maxsize: Union[int, Tuple[int, ...], None] = None,
) -> Union[List[Tuple[int, int]], list]:
    """Convert stacked list/tensor of starts end stops separated to list of (start, end) tuples.

    Args:
        x: (2+C, N) tensor, where C defines indices in dimensions of a segments for 3D activity tensors.
        maxsize: Optional max size. If None, use x.max().

    Returns:
        list of (start, end) tuples of shape (*, N, 2).
            note: (*) corresponds to C batched dimensions.
    """
    if segments.shape[0] in (0, 1):
        msg = f"Invalid argument shape {segments.shape=}. (expected first dim >= 2)"
        raise ValueError(msg)

    elif segments.shape[0] == 2:
        starts, ends = segments.tolist()
        return list(zip(starts, ends))

    if maxsize is None:
        num_elems = segments[0].max().item() + 1
        next_maxsize = None
    elif isinstance(maxsize, tuple):
        num_elems = maxsize[0]
        next_maxsize = maxsize[1:]
    else:
        num_elems = maxsize
        next_maxsize = None

    arange = torch.arange(num_elems)
    result = [
        segments_to_segments_list(
            segments[1:, ..., segments[0] == idx], maxsize=next_maxsize
        )
        for idx in arange
    ]
    return result


def segments_list_to_activity(
    segments_list: Union[List[Tuple[int, int]], Tensor, list],
    maxsize: Union[int, None] = None,
    *,
    device: DeviceLike = None,
) -> BoolTensor:
    """Convert list of (start, end) tuples to activity boolean tensor.

    Example
    =======
    ```python
    >>> segments = [(3, 6), (8, 9)]
    >>> segments_list_to_activity(segments)
    ... tensor([False, False, False, True, True, True, False, False, True])
    ```

    Args:
        segments_list: list of (start, end) tuples of shape (*, N, 2).
        maxsize: Optional max size. If None, use segments_list.max(). defaults to None.
        device: Optional output device. If None and segments_list is a tensor, it will use the same device. defaults to None.

    Returns:
        activity boolean tensor of shape (*, maxsize)
    """
    if device is None and isinstance(segments_list, Tensor):
        device = segments_list.device
    else:
        device = as_device(device)

    ndim = F.get_ndim(segments_list)
    if ndim == 1 and len(segments_list) == 0:
        if maxsize is None:
            num_elems = 0
        else:
            num_elems = maxsize

        activity = F.full((num_elems,), False, dtype=torch.bool, device=device)
        return activity  # type: ignore

    elif ndim == 2:
        segments_list = F.as_tensor(segments_list)
        return _segments_list_tensor_to_activity(segments_list, maxsize, device)

    elif (
        ndim > 2
        and F.get_shape(
            segments_list, return_indicator=True, return_default_on_invalid=True
        ).valid
    ):
        segments_list = F.as_tensor(segments_list)
        return _segments_list_tensor_to_activity(segments_list, maxsize, device)

    elif ndim > 2:
        activities = [
            segments_list_to_activity(segments_list_i, maxsize, device=device)  # type: ignore
            for segments_list_i in segments_list
        ]
        return pad_and_stack_rec(activities, False, device=device, dtype=torch.bool)  # type: ignore

    else:
        msg = f"Invalid argument ndim {ndim}. (expected ndim>=2 or (ndim == 1 and len == 0))"
        raise ValueError(msg)


def _segments_list_tensor_to_activity(
    segments_list: Tensor,
    maxsize: Optional[int],
    device: DeviceLike,
) -> BoolTensor:
    if device is None and isinstance(segments_list, Tensor):
        device = segments_list.device
    else:
        device = as_device(device)

    assert segments_list.ndim >= 2 and segments_list.shape[-1] == 2, (
        f"{segments_list.shape=}"
    )
    starts, ends = segments_list.permute(-1, *range(0, segments_list.ndim - 1))

    if maxsize is None:
        num_elems = ends.max().item()
    else:
        num_elems = maxsize

    unsqueeze_arange = tuple([None] * starts.ndim + [slice(None)])
    arange = F.arange(num_elems, device=device)[unsqueeze_arange]

    unsqueeze_bounds = tuple([slice(None)] * starts.ndim + [None])
    activity = (starts[unsqueeze_bounds] <= arange) & (arange < ends[unsqueeze_bounds])
    activity = activity.any(dim=-2)

    return activity  # type: ignore


def activity_to_segments_list(x: Tensor) -> Union[List[Tuple[int, int]], list]:
    segments = activity_to_segments(x)
    segments_lst = segments_to_segments_list(segments, x.shape[-1])
    return segments_lst


def segments_to_activity(x: Tensor, maxsize: Optional[int] = None) -> BoolTensor:
    """Convert stacked list/tensor of starts end stops separated to activity boolean tensor.

    Args:
        x: (*, 2, N) tensor
        maxsize: Optional max size. If None, use x.max().

    Returns:
        activity boolean tensor of shape (*, maxsize)
    """
    if maxsize is None:
        maxsize = int(x.max().item())
    segments_lst = segments_to_segments_list(x, maxsize)
    activity = segments_list_to_activity(segments_lst, maxsize)
    return activity



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchwrench.extras.numpy.saving import (  # noqa: F401
    dump_ndarray,
    dumps_ndarray,
    load_ndarray,
    loads_ndarray,
    read_ndarray,
    save_ndarray,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .slicer import DatasetSlicer, DatasetSlicerWrapper
from .tabular import T_ColIndex, T_RowIndex, TabularDataset
from .wrapper import (
    EmptyDataset,
    IterableSubset,
    IterableTransformWrapper,
    IterableWrapper,
    Subset,
    TransformWrapper,
    Wrapper,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torchwrench as tw
from torchwrench.utils.data.split import balanced_monolabel_split, random_split


class TestSplit(TestCase):
    def test_balanced_monolabel_split(self) -> None:
        num_classes = 5
        targets_indices = tw.randint(0, num_classes, (100,))
        splitted = balanced_monolabel_split(targets_indices, num_classes, [0.1, 0.2])

        assert len(splitted) == 2
        assert len(splitted[0]) <= 10
        assert len(splitted[1]) <= 20

    def test_random_split(self) -> None:
        splitted = random_split(100, [0.1, 0.2])

        assert len(splitted) == 2
        assert len(splitted[0]) <= 10
        assert len(splitted[1]) <= 20


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchwrench.core.packaging import _PANDAS_AVAILABLE

if not _PANDAS_AVAILABLE:
    from torchwrench.extras.pandas import _pandas_fallback as pd

else:
    import pandas as pd


__all__ = [
    "_PANDAS_AVAILABLE",
    "pd",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pythonwrench.functools import function_alias

from torchwrench.extras.hdf import HDFDataset, pack_to_hdf


@function_alias(pack_to_hdf)
def dump_hdf(*args, **kwargs): ...


@function_alias(HDFDataset)
def load_hdf(*args, **kwargs): ...


@function_alias(load_hdf)
def read_hdf(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchwrench.core.config import _REPLACE_MODULE_CLASSES
from torchwrench.nn.modules._mixins import EModule  # noqa: F401

if _REPLACE_MODULE_CLASSES:
    from torchwrench.nn.modules._mixins import EModule as Module  # noqa: F401
else:
    from torch.nn import Module  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import pythonwrench as pw

import torchwrench as tw
from torchwrench.extras.numpy import _NUMPY_AVAILABLE, np
from torchwrench.extras.pandas import _PANDAS_AVAILABLE, pd
from torchwrench.extras.speechbrain import _SPEECHBRAIN_AVAILABLE, DynamicItemDataset
from torchwrench.utils.data.dataset import TabularDataset


class TestTabularDataset(TestCase):
    def test_dict_list(self) -> None:
        data = {"a": list(range(5)), "b": list(range(5, 10))}
        ds = TabularDataset(data)

        assert len(ds) == 5
        assert ds.shape == (5, 2)
        assert ds.column_names == ds.keys() == ("a", "b")
        assert list(ds.values()) == [list(range(5)), list(range(5, 10))]
        assert ds.to_dict_list() == data
        assert ds.to_list_dict() == pw.dict_list_to_list_dict(data)

        # single_row=T, single_col=T, has_col=T
        assert ds[0, "a"] == 0

        # single_row=T, single_col=T, has_col=F
        # does not exist

        # single_row=T, single_col=F, has_col=T
        assert ds[2, ["a", "b"]] == [2, 7]

        # single_row=T, single_col=F, has_col=F
        assert ds[1] == {"a": 1, "b": 6}

        # single_row=F, single_col=T, has_col=T
        assert ds[[3, 4, 3], "b"] == [8, 9, 8]
        assert ds[:, "a"] == list(range(5))

        # single_row=F, single_col=T, has_col=F
        # does not exist

        # single_row=F, single_col=F, has_col=T
        assert ds[[3, 4, 3], ["b"]] == [[8, 9, 8]]
        assert ds[:, ["a"]] == [list(range(5))]

        # single_row=F, single_col=F, has_col=F
        assert ds[2:4] == [{"a": 2, "b": 7}, {"a": 3, "b": 8}]
        assert ds[[3, 4, 3]] == [{"a": 3, "b": 8}, {"a": 4, "b": 9}, {"a": 3, "b": 8}]

        if _NUMPY_AVAILABLE:
            expected = np.array([[0, 5], [1, 6], [2, 7], [3, 8], [4, 9]])
            assert tw.deep_equal(ds.to_numpy(), expected)

        if _PANDAS_AVAILABLE:
            assert tw.deep_equal(ds.to_dataframe(), pd.DataFrame(data))

    def test_ndarray(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        data: np.ndarray = np.random.rand(10, 3, 2)
        ds = TabularDataset(data)

        result_list_dict = ds.to_list_dict()
        expected_list_dict = [
            {j: data_ij for j, data_ij in enumerate(data_i)} for data_i in data
        ]
        assert tw.deep_equal(result_list_dict, expected_list_dict), (
            f"{result_list_dict=}; {expected_list_dict=}"
        )

        assert tw.deep_equal(ds.to_numpy(), data)
        assert tw.deep_equal(ds[0], data[0])
        assert tw.deep_equal(ds[:2], data[:2])
        assert tw.deep_equal(ds[1, 0], data[1, 0])
        assert tw.deep_equal(ds[5, 1, 0], data[5, 1, 0])
        assert tuple(ds.keys()) == tuple(ds.column_names) == (0, 1, 2)

    def test_dynamic_item_dataset(self) -> None:
        if not _SPEECHBRAIN_AVAILABLE:
            return None

        data = {f"{i}": {"string": i} for i in range(100)}
        ds = DynamicItemDataset(data)
        ds.set_output_keys(["string"])
        tabular = TabularDataset(ds)

        assert tabular[0] == {"string": 0}
        assert tabular[:3] == [{"string": i} for i in range(3)]
        assert tabular[3, "string"] == 3
        assert tabular[:3, "string"] == list(range(3))
        assert tabular[:3, ["string"]] == [[0], [1], [2]]

    def test_dynamic_column(self) -> None:
        data = {"a": list(range(5)), "b": list(range(5, 10))}
        ds = TabularDataset(data)

        def double(x):
            return x * 2

        ds.add_dynamic_column(double, requires=("a",), provides="c")

        assert ds[3] == {"a": 3, "b": 8, "c": 6}

        result = ds[4, "c"]
        expected = double(data["a"][4])
        assert result == expected, f"{result=}; {expected=}"

        mask = (
            tw.full((len(ds),), True)
            if not _NUMPY_AVAILABLE
            else np.full((len(ds),), True)
        )
        assert ds[mask] == ds.to_list_dict()

    def test_tensor(self) -> None:
        data = tw.rand(10, 2, 3)
        ds = TabularDataset(data)
        mask = tw.rand(data.shape[0]) > 0.5

        assert tw.deep_equal(ds.to_tensor(), data)
        assert tw.deep_equal(ds[mask], data[mask])

    def test_mapping(self) -> None:
        def triple(x):
            return x * 3

        data = {
            "a": ["a1", "a2", "a3"],
            "b": [1, 2, 3],
            "c": tw.arange(1, 4),
        }
        ds = TabularDataset(
            data,
            row_mapper={0: 1, 1: 1},
            col_mapper={"e": "a", "c": "c", "d": "d"},
            fns_list=[(("b",), "d", triple)],
        )

        assert len(ds) == 2
        assert ds.column_names == ("e", "c", "d")

        sample = ds[0]
        expected = {"e": "a2", "c": tw.as_tensor(2), "d": 6}
        assert tw.deep_equal(sample, expected), f"{sample=}; {expected=}"

        datadict = ds.to_dict_list()
        expected = {
            "e": ["a2", "a2"],
            "c": [tw.as_tensor(2), tw.as_tensor(2)],
            "d": [6, 6],
        }
        assert tw.deep_equal(datadict, expected)

        mask = [False] * len(ds)
        sample = ds[mask]
        assert sample == [], f"{sample=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from io import StringIO, TextIOBase
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
    get_args,
    overload,
)

from pythonwrench._core import _setup_output_fpath
from pythonwrench.csv import dump_csv as _dump_csv_base
from pythonwrench.csv import load_csv as _load_csv_base
from pythonwrench.functools import function_alias
from pythonwrench.warnings import warn_once

from torchwrench.core.packaging import _PANDAS_AVAILABLE
from torchwrench.extras.pandas import pd
from torchwrench.serialization.common import as_builtin

OrientExtended = Literal["list", "dict", "dataframe", "auto"]
CSVBackend = Literal["csv", "pandas", "auto"]


def dump_csv(
    data: Union[Iterable[Mapping[str, Any]], Mapping[str, Iterable[Any]], Iterable],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    backend: CSVBackend = "auto",
    header: Union[bool, Literal["auto"]] = "auto",
    **csv_backend_kwds,
) -> str:
    """Dump content to csv format."""
    if backend == "auto":
        if isinstance(data, pd.DataFrame):
            backend = "pandas"
        else:
            backend = "csv"

    if backend == "csv":
        return _dump_csv_base(
            data,
            fpath,
            overwrite=overwrite,
            make_parents=make_parents,
            to_builtins=to_builtins,
            header=header,
            **csv_backend_kwds,
        )

    elif backend == "pandas":
        if to_builtins:
            if isinstance(data, pd.DataFrame):
                msg = f"Inconsistent combinaison of arguments: {to_builtins=}, {backend=} and {type(data)=}."
                warn_once(msg)
            data = as_builtin(data)

        header = header if header != "auto" else True
        return _dump_csv_with_pandas(
            data,
            fpath,
            overwrite=overwrite,
            make_parents=make_parents,
            header=header,
            **csv_backend_kwds,
        )

    else:
        msg = f"Invalid argument {backend=}. (expected one of {get_args(CSVBackend)})"
        raise ValueError(msg)


@function_alias(dump_csv)
def dumps_csv(*args, **kwargs): ...


@function_alias(dump_csv)
def save_csv(*args, **kwargs): ...


@overload
def load_csv(
    fpath: Union[str, Path, TextIOBase],
    /,
    *,
    orient: Literal["list"] = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "auto",
    # CSV reader kwargs
    delimiter: Optional[str] = None,
    **csv_backend_kwds,
) -> List[Dict[str, Any]]: ...


@overload
def load_csv(
    fpath: Union[str, Path, TextIOBase],
    /,
    *,
    orient: Literal["dict"],
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "auto",
    # CSV reader kwargs
    delimiter: Optional[str] = None,
    **csv_backend_kwds,
) -> Dict[str, List[Any]]: ...


@overload
def load_csv(
    fpath: Union[str, Path, TextIOBase],
    /,
    *,
    orient: Literal["dataframe"],
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "auto",
    # CSV reader kwargs
    delimiter: Optional[str] = None,
    **csv_backend_kwds,
) -> pd.DataFrame: ...


def load_csv(
    fpath: Union[str, Path, TextIOBase],
    /,
    *,
    orient: OrientExtended = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    backend: CSVBackend = "auto",
    # CSV reader kwargs
    delimiter: Optional[str] = None,
    **csv_backend_kwds,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]], pd.DataFrame]:
    """Load CSV file using CSV or pandas backend."""
    if backend == "auto":
        if _PANDAS_AVAILABLE:
            backend = "pandas"
        else:
            backend = "csv"

    if orient == "auto":
        if _PANDAS_AVAILABLE:
            orient = "dataframe"
        else:
            orient = "list"

    if backend == "csv":
        if orient == "dataframe":
            if not _PANDAS_AVAILABLE:
                msg = f"Invalid argument {backend=} without pandas installed."
                raise ValueError(msg)
            backend_orient = "dict"
        else:
            backend_orient = orient

        result = _load_csv_base(
            fpath,
            orient=backend_orient,
            header=header,
            comment_start=comment_start,
            strip_content=strip_content,
            delimiter=delimiter,
            **csv_backend_kwds,
        )

        if orient == "dataframe":
            result = pd.DataFrame(result)

    elif backend == "pandas":
        result = _load_csv_with_pandas(
            fpath,
            orient=orient,
            header=header,
            comment_start=comment_start,
            strip_content=strip_content,
            delimiter=delimiter,
            **csv_backend_kwds,
        )

    else:
        msg = f"Invalid argument {backend=}. (expected one of {get_args(CSVBackend)})"
        raise ValueError(msg)

    return result


@function_alias(load_csv)
def loads_csv(*args, **kwargs): ...


@function_alias(load_csv)
def read_csv(*args, **kwargs): ...


def _dump_csv_with_pandas(
    data: Union[Iterable[Mapping[str, Any]], Mapping[str, Iterable[Any]], pd.DataFrame],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    **kwargs,
) -> str:
    backend = "pandas"
    if not _PANDAS_AVAILABLE:
        msg = f"Invalid argument {backend=} without pandas installed."
        raise ValueError(msg)

    df = pd.DataFrame(data)  # type: ignore

    # set index to False by default
    kwargs.setdefault("index", False)

    file = StringIO()
    df.to_csv(file, **kwargs)
    content = file.getvalue()
    file.close()

    fpath = _setup_output_fpath(fpath, overwrite, make_parents)
    if fpath is not None:
        fpath.write_text(content)

    return content


def _load_csv_with_pandas(
    fpath: Union[str, Path, TextIOBase],
    /,
    *,
    orient: OrientExtended = "list",
    header: bool = True,
    comment_start: Optional[str] = None,
    strip_content: bool = False,
    # Backend kwargs
    delimiter: Optional[str] = None,
    **csv_backend_kwds,
) -> Union[List[Dict[str, Any]], Dict[str, List[Any]], pd.DataFrame]:
    backend = "pandas"

    if not _PANDAS_AVAILABLE:
        msg = f"Invalid argument {backend=} without pandas installed."
        raise ValueError(msg)

    if strip_content:
        msg = f"Invalid argument {strip_content=} with {backend=}."
        raise ValueError(msg)

    if comment_start is not None:
        msg = f"Invalid argument {comment_start=} with {backend=}."
        raise ValueError(msg)

    if len(csv_backend_kwds) > 0:
        msg = f"Invalid arguments {csv_backend_kwds=} with {backend=}."
        raise ValueError(msg)

    df = pd.read_csv(fpath, delimiter=delimiter)

    if orient == "list":
        return df.to_dict("records")  # type: ignore
    elif orient == "dict":
        return df.to_dict("list")  # type: ignore
    elif orient in ("pandas", "auto"):
        return df
    else:
        msg = (
            f"Invalid argument {orient=}. (expected one of {get_args(OrientExtended)})"
        )
        raise ValueError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tensor subclasses for typing and instance checks.

Note: torchwrench.FloatTensor != torch.FloatTensor which is caused by `issubclass(torchwrench.FloatTensor, torch.FloatTensor) is False` because torch.FloatTensor cannot be subclassed

Here is an overview of the valid tensor subclasses tree:

                                                  Tensor
                                                    |
                  +---------------------------------+---------------------------------+
                  |                                                                   |
        ComplexFloatingTensor                                                     RealTensor
                  |                                                                   |
     +------------+------------+                          +---------------------------+----------------------------+
     |            |            |                          |                                                        |
CHalfTensor CFloatTensor CDoubleTensor              FloatingTensor                                           IntegralTensor
   (c32)        (c64)       (c128)                        |                                                        |
                                              +-----------+-----------+                         +------------------+------------------+
                                              |           |           |                         |                                     |
                                          HalfTensor FloatTensor DoubleTensor          SignedIntegerTensor                  UnsignedIntegerTensor
                                            (f16)       (f32)       (f64)                       |                                     |
                                                                              +-----------+-----+-----+-----------+             +-----+-----+
                                                                              |           |           |           |             |           |
                                                                          CharTensor  ShortTensor  IntTensor  LongTensor   ByteTensor   BoolTensor
                                                                             (i8)       (i16)       (i32)       (i64)         (u8)       (bool)
"""

import builtins
from typing import (
    Any,
    ClassVar,
    Dict,
    Final,
    Generic,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
    overload,
)

import torch
from pythonwrench import BuiltinNumber
from torch._C import _TensorMeta
from torch.types import Device
from typing_extensions import TypeVar

from torchwrench.core.dtype_enum import DTypeEnum
from torchwrench.core.make import DeviceLike, DTypeLike, as_device, as_dtype

_bool = builtins.bool
_float = builtins.float
_int = builtins.int

_DEFAULT_T_DTYPE = Literal[None]
_DEFAULT_T_NDIM = _int
_DEFAULT_T_FLOATING = _bool
_DEFAULT_T_COMPLEX = _bool
_DEFAULT_T_SIGNED = _bool

T_BuiltinNumber = TypeVar("T_BuiltinNumber", bound=BuiltinNumber, covariant=True)
T_Tensor = TypeVar("T_Tensor", bound="_TensorNDBase", contravariant=True)

T_DType = TypeVar("T_DType", bound=Union["DTypeEnum", None])
T_NDim = TypeVar("T_NDim", bound=_int)
T_Floating = TypeVar("T_Floating", bound=_bool)
T_Complex = TypeVar("T_Complex", bound=_bool)
T_Signed = TypeVar("T_Signed", bound=_bool)

_TORCH_BASE_CLASSES: Final[Dict[str, Type]] = {
    "float32": torch.FloatTensor,
    "float": torch.FloatTensor,
    "float64": torch.DoubleTensor,
    "double": torch.DoubleTensor,
    "float16": torch.HalfTensor,
    "half": torch.HalfTensor,
    "int16": torch.ShortTensor,
    "short": torch.ShortTensor,
    "int32": torch.IntTensor,
    "int": torch.IntTensor,
    "int64": torch.LongTensor,
    "long": torch.LongTensor,
    "bool": torch.BoolTensor,
    "byte": torch.ByteTensor,
    "char": torch.CharTensor,
    "int8": torch.CharTensor,
}


class _GenericsValues(NamedTuple):
    dtype: Optional[torch.dtype] = None
    ndim: Optional[_int] = None
    is_floating_point: Optional[_bool] = None
    is_complex: Optional[_bool] = None
    is_signed: Optional[_bool] = None

    def is_compatible_with_tensor(self, tensor: torch.Tensor) -> _bool:
        if self.ndim is not None and self.ndim != tensor.ndim:
            return False
        else:
            return self.is_compatible_with_dtype(tensor.dtype)

    def is_compatible_with_dtype(self, dtype: torch.dtype) -> _bool:
        if self.dtype is not None:
            return self.dtype == dtype

        for self_attr, dtype_attr in zip(
            self[2:5], (dtype.is_floating_point, dtype.is_complex, dtype.is_signed)
        ):
            if self_attr is not None and self_attr != dtype_attr:
                return False
        return True

    def is_compatible_with_generic(self, other: "_GenericsValues") -> _bool:
        for self_attr, other_attr in zip(self, other):
            if self_attr is not None and (
                other_attr is None or self_attr != other_attr
            ):
                return False
        return True


def _get_generics(
    cls: Union["_TensorNDMeta", "_TensorNDBase"],
) -> _GenericsValues:
    if hasattr(cls, "__orig_class__"):
        orig = cls.__orig_class__  # type: ignore
        if orig.__origin__ is not _TensorNDMeta:
            return _GenericsValues()

        generic_args = orig.__args__  # type: ignore
        generic_args = generic_args  # currently only check dtype and ndim

    elif get_origin(cls) is not None:
        args = get_args(cls)
        generic_args = args + _GenericsValues()[len(args) :]
    else:
        return _GenericsValues()

    assert len(generic_args) == 6

    t_dtype = generic_args[0]
    if t_dtype is not _DEFAULT_T_DTYPE:
        enum_dtype: DTypeEnum = t_dtype.__args__[0]
        dtype = enum_dtype.dtype
    else:
        dtype = None

    t_ndim = generic_args[1]
    if t_ndim is not _DEFAULT_T_NDIM:
        ndim = t_ndim.__args__[0]
    else:
        ndim = None

    t_floating = generic_args[3]
    if t_floating is not _DEFAULT_T_FLOATING:
        floating = t_floating.__args__[0]
    else:
        floating = None

    t_complex = generic_args[4]
    if t_complex is not _DEFAULT_T_COMPLEX:
        complex_ = t_complex.__args__[0]
    else:
        complex_ = None

    t_signed = generic_args[5]
    if t_signed is not _DEFAULT_T_SIGNED:
        signed = t_signed.__args__[0]
    else:
        signed = None

    return _GenericsValues(dtype, ndim, floating, complex_, signed)


class _TensorNDMeta(
    Generic[T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed],
    _TensorMeta,
):
    """Tensor metaclass with redefined instance check based on generic properties.

    Generic attributes are: T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed
    """

    def __instancecheck__(self, instance: Any) -> _bool:
        """Called method to check isinstance(instance, self)"""
        if not isinstance(instance, torch.Tensor):
            return False

        gen = _get_generics(self)
        return gen.is_compatible_with_tensor(instance)

    def __subclasscheck__(self, subclass: Any) -> _bool:
        """Called method to check issubclass(subclass, cls)"""
        self_generics = _get_generics(self)
        other_generics = _get_generics(subclass)
        return self_generics.is_compatible_with_generic(other_generics)


class _TensorNDBase(
    Generic[T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed],
    torch.Tensor,
):
    """Tensor base class with redefined constructor and overloaded methods.

    Generic attributes are: T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed
    """

    _DEFAULT_DTYPE: ClassVar[Optional[DTypeEnum]] = None

    @overload
    def __new__(
        cls: Type[T_Tensor],
        *dims: _int,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
        memory_format: Union[torch.memory_format, None] = None,
        out: Union[torch.Tensor, None] = None,
        layout: Union[torch.layout, None] = None,
        pin_memory: Union[_bool, None] = False,
        requires_grad: Union[_bool, None] = False,
    ) -> T_Tensor: ...

    @overload
    def __new__(
        cls: Type[T_Tensor],
        data: Union[T_BuiltinNumber, Sequence],
        /,
        *,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
    ) -> T_Tensor: ...

    def __new__(
        cls: Type[T_Tensor],
        *args: Any,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
        memory_format: Union[torch.memory_format, None] = None,
        out: Union[torch.Tensor, None] = None,
        layout: Union[torch.layout, None] = None,
        pin_memory: Union[_bool, None] = False,
        requires_grad: Union[_bool, None] = False,
    ) -> T_Tensor:
        dtype = as_dtype(dtype)
        device = as_device(device)

        gen = _get_generics(cls)  # type: ignore
        cls_dtype = gen.dtype
        cls_ndim = gen.ndim

        # Sanity checks for dtype
        if dtype is None:
            if cls_dtype is not None:
                dtype = cls_dtype
            elif cls._DEFAULT_DTYPE is not None:
                dtype = cls._DEFAULT_DTYPE.dtype

        elif cls_dtype is None:
            if not gen.is_compatible_with_dtype(dtype):
                msg = f"Invalid argument {dtype=} for {cls.__name__}. (expected a dtype with (is_floating_point={gen.is_floating_point}, is_complex={gen.is_complex}, is_signed={gen.is_signed}))"
                raise ValueError(msg)

        elif dtype == cls_dtype:
            pass
        else:
            msg = (
                f"Invalid argument {dtype=} for {cls.__name__}. (expected {cls_dtype})"
            )
            raise ValueError(msg)

        # Sanity checks for data and ndim
        is_int_args = all(isinstance(arg, _int) for arg in args)
        if cls_ndim is None:
            if len(args) == 0:
                size = (0,)
                data = None
            elif is_int_args:
                size = args
                data = None
            elif len(args) == 1 and not isinstance(args[0], _int):
                size = None
                data = args[0]
            else:
                msg = f"Invalid arguments {args=}. (expected only ints or one sequence of data)"
                raise ValueError(msg)

        elif cls_ndim == len(args) and is_int_args:
            size = args
            data = None
        elif len(args) == 1:
            size = None
            data = args[0]
        elif len(args) == 0:
            size = [0] * cls_ndim
            data = None
        else:
            msg = f"Invalid arguments {args=}. (expected {cls_ndim} ints or one sequence of data)"
            raise ValueError(msg)
        del args

        # Supports older PyTorch versions
        if layout is None:
            layout = torch.strided
        if pin_memory is None:
            pin_memory = False
        if requires_grad is None:
            requires_grad = False

        if data is not None:
            result = torch.as_tensor(
                data=data,
                dtype=dtype,  # type: ignore
                device=device,
            )
            if cls_ndim is not None and result.ndim != cls_ndim:
                msg = f"Invalid number of dimension(s) for argument data in {cls.__name__}. (found {result.ndim} but expected {cls_ndim})"
                raise ValueError(msg)
            return result  # type: ignore

        elif size is not None:
            return torch.empty(
                size,
                dtype=dtype,  # type: ignore
                device=device,
                memory_format=memory_format,
                out=out,
                layout=layout,
                pin_memory=pin_memory,
                requires_grad=requires_grad,
            )  # type: ignore

        else:
            msg = f"Internal error: found {data=} and {size=} in {cls.__name__}."
            raise RuntimeError(msg)

    @overload
    def __init__(
        self,
        *dims: _int,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
        memory_format: Union[torch.memory_format, None] = None,
        out: Union[torch.Tensor, None] = None,
        layout: Union[torch.layout, None] = None,
        pin_memory: Union[_bool, None] = False,
        requires_grad: Union[_bool, None] = False,
    ) -> None: ...

    @overload
    def __init__(
        self,
        data: Union[T_BuiltinNumber, Sequence],
        /,
        *,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
    ) -> None: ...

    def __init__(
        self,
        *args: Any,
        dtype: DTypeLike = None,
        device: DeviceLike = None,
        memory_format: Union[torch.memory_format, None] = None,
        out: Union[torch.Tensor, None] = None,
        layout: Union[torch.layout, None] = None,
        pin_memory: Union[_bool, None] = False,
        requires_grad: Union[_bool, None] = False,
    ) -> None: ...

    @overload
    def __eq__(self, other: Any) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor1D", idx: _int, /) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor2D", idx: _int, /) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor3D", idx: _int, /) -> "Tensor2D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor0D", idx: None, /) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor1D", idx: None, /) -> "Tensor2D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor2D", idx: None, /) -> "Tensor3D":  # type: ignore
        ...

    @overload
    def __getitem__(self: "Tensor3D", idx: None, /) -> "Tensor":  # type: ignore
        ...

    @overload
    def __getitem__(self: T_Tensor, sl: slice, /) -> T_Tensor: ...

    @overload
    def __getitem__(self, *args) -> "Tensor": ...

    @overload
    def __ne__(self, other: Any) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def abs(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def absolute(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def acos(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def all(self, dim: Literal[None] = None) -> "BoolTensor0D":  # type: ignore
        ...

    @overload
    def all(  # type: ignore
        self,
        dim: Union[_int, Tuple[_int, ...]],
        keepdim: _bool = False,
    ) -> "BoolTensor": ...

    @overload
    def any(self, dim: Literal[None] = None) -> "BoolTensor0D":  # type: ignore
        ...

    @overload
    def any(  # type: ignore
        self,
        dim: Union[_int, Tuple[_int, ...]],
        keepdim: _bool = False,
    ) -> "BoolTensor": ...

    @overload
    def bool(self: T_Tensor) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def contiguous(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def cpu(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def cuda(self: T_Tensor) -> T_Tensor:  # type: ignore
        ...

    @overload
    def double(self) -> "DoubleTensor":  # type: ignore
        ...

    @overload
    def eq(self: T_Tensor, other: Union[torch.Tensor, BuiltinNumber]) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def equal(self: T_Tensor, other: torch.Tensor) -> _bool:  # type: ignore
        ...

    @overload
    def float(self) -> "FloatTensor":  # type: ignore
        ...

    @overload
    def half(self) -> "HalfTensor":  # type: ignore
        ...

    @overload
    def int(self) -> "IntTensor":  # type: ignore
        ...

    @overload
    def is_complex(self) -> T_Complex:  # type: ignore
        ...

    @overload
    def is_floating_point(self) -> T_Floating:  # type: ignore
        ...

    @overload
    def is_signed(self) -> T_Signed:  # type: ignore
        ...

    @overload
    def isfinite(self) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def isinf(self) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def isnan(self) -> "BoolTensor":  # type: ignore
        ...

    @overload
    def item(self) -> T_BuiltinNumber:  # type: ignore
        ...

    @overload
    def long(self) -> "LongTensor":  # type: ignore
        ...

    @overload
    def mean(self, dim: Literal[None] = None) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def mean(self: "Tensor0D", dim: _int) -> "Tensor0D": ...

    @overload
    def mean(self: "Tensor1D", dim: _int) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def mean(self: "Tensor2D", dim: _int) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def mean(self: "Tensor3D", dim: _int) -> "Tensor2D":  # type: ignore
        ...

    @overload
    def mean(self, dim: _int) -> "Tensor":  # type: ignore
        ...

    @overload
    def reshape(self, size: Tuple[()]) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def reshape(self, size: Tuple[_int]) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def reshape(self, size: Tuple[_int, _int]) -> "Tensor2D": ...

    @overload
    def reshape(self, size: Tuple[_int, _int, _int]) -> "Tensor3D": ...

    @overload
    def reshape(self, size: Tuple[_int, ...]) -> "Tensor": ...

    @overload
    def reshape(self, size0: _int) -> "Tensor1D": ...

    @overload
    def reshape(self, size0: _int, size1: _int) -> "Tensor2D": ...

    @overload
    def reshape(self, size0: _int, size1: _int, size2: _int) -> "Tensor3D":  # type: ignore
        ...

    @overload
    def short(self) -> "ShortTensor":  # type: ignore
        ...

    @overload
    def squeeze(self, dim: Optional[_int] = None) -> "Tensor":  # type: ignore
        ...

    @overload
    def sum(self, dim: Literal[None] = None) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def sum(self: "Tensor1D", dim: _int) -> "Tensor0D": ...

    @overload
    def sum(self: "Tensor2D", dim: _int) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def sum(self: "Tensor3D", dim: _int) -> "Tensor2D":  # type: ignore
        ...

    @overload
    def sum(self, dim: Optional[_int] = None) -> "Tensor":  # type: ignore
        ...

    @overload
    def to(  # type: ignore
        self: T_Tensor,
        dtype: Optional[torch.dtype] = None,
        non_blocking: _bool = False,
        copy: _bool = False,
        *,
        memory_format: Optional[torch.memory_format] = None,
    ) -> T_Tensor: ...

    @overload
    def to(
        self: T_Tensor,
        device: Device = None,
        dtype: Optional[torch.dtype] = None,
        non_blocking: _bool = False,
        copy: _bool = False,
        *,
        memory_format: Optional[torch.memory_format] = None,
    ) -> T_Tensor: ...

    @overload
    def to(  # type: ignore
        self,
        other: T_Tensor,
        non_blocking: _bool = False,
        copy: _bool = False,
        *,
        memory_format: Optional[torch.memory_format] = None,
    ) -> T_Tensor: ...

    @overload
    def tolist(self) -> Any:  # type: ignore
        ...

    @overload
    def unsqueeze(self: "Tensor0D", dim: _int) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def unsqueeze(self: "Tensor1D", dim: _int) -> "Tensor2D":  # type: ignore
        ...

    @overload
    def unsqueeze(self: "Tensor2D", dim: _int) -> "Tensor3D":  # type: ignore
        ...

    @overload
    def unsqueeze(self, dim: _int) -> "Tensor":  # type: ignore
        ...

    @overload
    def view(self, size: Tuple[()]) -> "Tensor0D":  # type: ignore
        ...

    @overload
    def view(self, size: Tuple[_int]) -> "Tensor1D":  # type: ignore
        ...

    @overload
    def view(self, size: Tuple[_int, _int]) -> "Tensor2D": ...

    @overload
    def view(self, size: Tuple[_int, _int, _int]) -> "Tensor3D": ...

    @overload
    def view(self, size: Tuple[_int, ...]) -> "Tensor": ...

    @overload
    def view(self, size0: _int) -> "Tensor1D": ...

    @overload
    def view(self, size0: _int, size1: _int) -> "Tensor2D": ...

    @overload
    def view(self, size0: _int, size1: _int, size2: _int) -> "Tensor3D": ...

    @overload
    def view(self, *size: _int) -> "Tensor":  # type: ignore
        ...

    @overload
    def view(self, dtype: torch.dtype) -> "Tensor":  # type: ignore
        ...

    ndim: T_NDim  # type: ignore

    __eq__ = torch.Tensor.__eq__  # noqa: F811  # type: ignore
    __getitem__ = torch.Tensor.__getitem__  # noqa: F811  # type: ignore
    __ne__ = torch.Tensor.__ne__  # noqa: F811  # type: ignore
    abs = torch.Tensor.abs  # noqa: F811  # type: ignore
    absolute = torch.Tensor.absolute  # noqa: F811  # type: ignore
    acos = torch.Tensor.acos  # noqa: F811  # type: ignore
    all = torch.Tensor.all  # noqa: F811  # type: ignore
    any = torch.Tensor.any  # noqa: F811  # type: ignore
    bool = torch.Tensor.bool  # noqa: F811  # type: ignore
    contiguous = torch.Tensor.contiguous  # noqa: F811  # type: ignore
    cpu = torch.Tensor.cpu  # noqa: F811  # type: ignore
    cuda = torch.Tensor.cuda  # noqa: F811  # type: ignore
    double = torch.Tensor.double  # noqa: F811  # type: ignore
    eq = torch.Tensor.eq  # noqa: F811  # type: ignore
    equal = torch.Tensor.equal  # noqa: F811  # type: ignore
    float = torch.Tensor.float  # noqa: F811  # type: ignore
    half = torch.Tensor.half  # noqa: F811  # type: ignore
    is_complex = torch.Tensor.is_complex  # noqa: F811  # type: ignore
    is_floating_point = torch.Tensor.is_floating_point  # noqa: F811  # type: ignore
    is_signed = torch.Tensor.is_signed  # noqa: F811  # type: ignore
    isfinite = torch.Tensor.isfinite  # noqa: F811  # type: ignore
    isinf = torch.Tensor.isinf  # noqa: F811  # type: ignore
    isnan = torch.Tensor.isnan  # noqa: F811  # type: ignore
    int = torch.Tensor.int  # noqa: F811  # type: ignore
    item = torch.Tensor.item  # noqa: F811  # type: ignore
    long = torch.Tensor.long  # noqa: F811  # type: ignore
    mean = torch.Tensor.mean  # noqa: F811  # type: ignore
    reshape = torch.Tensor.reshape  # noqa: F811  # type: ignore
    short = torch.Tensor.short  # noqa: F811  # type: ignore
    squeeze = torch.Tensor.squeeze  # noqa: F811  # type: ignore
    sum = torch.Tensor.sum  # noqa: F811  # type: ignore
    to = torch.Tensor.to  # noqa: F811  # type: ignore
    tolist = torch.Tensor.tolist  # noqa: F811  # type: ignore
    unsqueeze = torch.Tensor.unsqueeze  # noqa: F811  # type: ignore
    view = torch.Tensor.view  # noqa: F811  # type: ignore


# # TODO: rm
# class TTensor(
#     Generic[T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed],
#     _TensorNDBase[T_DType, T_NDim, T_BuiltinNumber, T_Floating, T_Complex, T_Signed],
#     metaclass=_TensorNDMeta,
# ): ...


class Tensor(
    _TensorNDBase[Literal[None], _int, BuiltinNumber, _bool, _bool, _bool],
    metaclass=_TensorNDMeta[Literal[None], _int, BuiltinNumber, _bool, _bool, _bool],
): ...


class Tensor0D(
    _TensorNDBase[Literal[None], Literal[0], BuiltinNumber, _bool, _bool, _bool],
    metaclass=_TensorNDMeta[
        Literal[None], Literal[0], BuiltinNumber, _bool, _bool, _bool
    ],
): ...


class Tensor1D(
    _TensorNDBase[Literal[None], Literal[1], BuiltinNumber, _bool, _bool, _bool],
    metaclass=_TensorNDMeta[
        Literal[None], Literal[1], BuiltinNumber, _bool, _bool, _bool
    ],
): ...


class Tensor2D(
    _TensorNDBase[Literal[None], Literal[2], BuiltinNumber, _bool, _bool, _bool],
    metaclass=_TensorNDMeta[
        Literal[None], Literal[2], BuiltinNumber, _bool, _bool, _bool
    ],
): ...


class Tensor3D(
    _TensorNDBase[Literal[None], Literal[3], BuiltinNumber, _bool, _bool, _bool],
    metaclass=_TensorNDMeta[
        Literal[None], Literal[3], BuiltinNumber, _bool, _bool, _bool
    ],
): ...


# ----------------------------------------
# Concrete classes
# ----------------------------------------


class BoolTensor(
    _TensorNDBase[
        Literal[DTypeEnum.bool],
        _int,
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.bool],
        _int,
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
): ...


class BoolTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.bool],
        Literal[0],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.bool],
        Literal[0],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> _bool:
        return super().tolist()  # type: ignore


class BoolTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.bool],
        Literal[1],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.bool],
        Literal[1],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[_bool]:
        return super().tolist()  # type: ignore


class BoolTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.bool],
        Literal[2],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.bool],
        Literal[2],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[List[_bool]]:
        return super().tolist()  # type: ignore


class BoolTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.bool],
        Literal[3],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.bool],
        Literal[3],
        _bool,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[List[List[_bool]]]:
        return super().tolist()  # type: ignore


class ByteTensor(
    _TensorNDBase[
        Literal[DTypeEnum.uint8],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.uint8],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
): ...


class ByteTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.uint8],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.uint8],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> _int:
        return super().tolist()  # type: ignore


class ByteTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.uint8],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.uint8],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[_int]:
        return super().tolist()  # type: ignore


class ByteTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.uint8],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.uint8],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[List[_int]]:
        return super().tolist()  # type: ignore


class ByteTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.uint8],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.uint8],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    def tolist(self) -> List[List[List[_int]]]:
        return super().tolist()  # type: ignore


class CharTensor(
    _TensorNDBase[
        Literal[DTypeEnum.int8],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int8],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
): ...


class CharTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.int8],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int8],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _int:
        return super().tolist()  # type: ignore


class CharTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.int8],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int8],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_int]:
        return super().tolist()  # type: ignore


class CharTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.int8],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int8],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_int]]:
        return super().tolist()  # type: ignore


class CharTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.int8],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int8],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_int]]]:
        return super().tolist()  # type: ignore


class DoubleTensor(
    _TensorNDBase[
        Literal[DTypeEnum.double],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.double],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
): ...


class DoubleTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.double],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.double],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _float:
        return super().tolist()  # type: ignore


class DoubleTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.double],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.double],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_float]:
        return super().tolist()  # type: ignore


class DoubleTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.double],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.double],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_float]]:
        return super().tolist()  # type: ignore


class DoubleTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.double],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.double],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_float]]]:
        return super().tolist()  # type: ignore


class FloatTensor(
    _TensorNDBase[
        Literal[DTypeEnum.float],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.float],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
): ...


class FloatTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.float],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.float],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
): ...


class FloatTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.float],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.float],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_float]:
        return super().tolist()  # type: ignore


class FloatTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.float],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.float],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_float]]:
        return super().tolist()  # type: ignore


class FloatTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.float],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.float],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_float]]]:
        return super().tolist()  # type: ignore


class HalfTensor(
    _TensorNDBase[
        Literal[DTypeEnum.half],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.half],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
): ...


class HalfTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.half],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.half],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _float:
        return super().tolist()  # type: ignore


class HalfTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.half],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.half],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_float]:
        return super().tolist()  # type: ignore


class HalfTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.half],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.half],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_float]]:
        return super().tolist()  # type: ignore


class HalfTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.half],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.half],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_float]]]:
        return super().tolist()  # type: ignore


class IntTensor(
    _TensorNDBase[
        Literal[DTypeEnum.int],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
): ...


class IntTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.int],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _int:
        return super().tolist()  # type: ignore


class IntTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.int],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_int]:
        return super().tolist()  # type: ignore


class IntTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.int],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_int]]:
        return super().tolist()  # type: ignore


class IntTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.int],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.int],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_int]]]:
        return super().tolist()  # type: ignore


class LongTensor(
    _TensorNDBase[
        Literal[DTypeEnum.long],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.long],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
): ...


class LongTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.long],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.long],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _int:
        return super().tolist()  # type: ignore


class LongTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.long],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.long],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_int]:
        return super().tolist()  # type: ignore


class LongTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.long],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.long],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_int]]:
        return super().tolist()  # type: ignore


class LongTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.long],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.long],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_int]]]:
        return super().tolist()  # type: ignore


class ShortTensor(
    _TensorNDBase[
        Literal[DTypeEnum.short],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.short],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
): ...


class ShortTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.short],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.short],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> _int:
        return super().tolist()  # type: ignore


class ShortTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.short],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.short],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[_int]:
        return super().tolist()  # type: ignore


class ShortTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.short],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.short],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[_int]]:
        return super().tolist()  # type: ignore


class ShortTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.short],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.short],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[_int]]]:
        return super().tolist()  # type: ignore


class CFloatTensor(
    _TensorNDBase[
        Literal[DTypeEnum.cfloat],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cfloat],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
): ...


class CFloatTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.cfloat],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cfloat],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> complex:
        return super().tolist()  # type: ignore


class CFloatTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.cfloat],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cfloat],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[complex]:
        return super().tolist()  # type: ignore


class CFloatTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.cfloat],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cfloat],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[complex]]:
        return super().tolist()  # type: ignore


class CFloatTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.cfloat],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cfloat],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[complex]]]:
        return super().tolist()  # type: ignore


if hasattr(torch, "complex32"):

    class CHalfTensor(
        _TensorNDBase[
            Literal[DTypeEnum.chalf],
            _int,
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
        metaclass=_TensorNDMeta[
            Literal[DTypeEnum.chalf],
            _int,
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
    ): ...

    class CHalfTensor0D(
        _TensorNDBase[
            Literal[DTypeEnum.chalf],
            Literal[0],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
        metaclass=_TensorNDMeta[
            Literal[DTypeEnum.chalf],
            Literal[0],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
    ):
        def tolist(self) -> complex:
            return super().tolist()  # type: ignore

    class CHalfTensor1D(
        _TensorNDBase[
            Literal[DTypeEnum.chalf],
            Literal[1],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
        metaclass=_TensorNDMeta[
            Literal[DTypeEnum.chalf],
            Literal[1],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
    ):
        def tolist(self) -> List[complex]:
            return super().tolist()  # type: ignore

    class CHalfTensor2D(
        _TensorNDBase[
            Literal[DTypeEnum.chalf],
            Literal[2],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
        metaclass=_TensorNDMeta[
            Literal[DTypeEnum.chalf],
            Literal[2],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
    ):
        def tolist(self) -> List[List[complex]]:
            return super().tolist()  # type: ignore

    class CHalfTensor3D(
        _TensorNDBase[
            Literal[DTypeEnum.chalf],
            Literal[3],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
        metaclass=_TensorNDMeta[
            Literal[DTypeEnum.chalf],
            Literal[3],
            complex,
            Literal[False],
            Literal[True],
            Literal[True],
        ],
    ):
        def tolist(self) -> List[List[List[complex]]]:
            return super().tolist()  # type: ignore


class CDoubleTensor(
    _TensorNDBase[
        Literal[DTypeEnum.cdouble],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cdouble],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
): ...


class CDoubleTensor0D(
    _TensorNDBase[
        Literal[DTypeEnum.cdouble],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cdouble],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> complex:
        return super().tolist()  # type: ignore


class CDoubleTensor1D(
    _TensorNDBase[
        Literal[DTypeEnum.cdouble],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cdouble],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[complex]:
        return super().tolist()  # type: ignore


class CDoubleTensor2D(
    _TensorNDBase[
        Literal[DTypeEnum.cdouble],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cdouble],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[complex]]:
        return super().tolist()  # type: ignore


class CDoubleTensor3D(
    _TensorNDBase[
        Literal[DTypeEnum.cdouble],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[DTypeEnum.cdouble],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    def tolist(self) -> List[List[List[complex]]]:
        return super().tolist()  # type: ignore


# ----------------------------------------
# Intermediate classes
# ----------------------------------------


class ComplexFloatingTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    """Intermediate class for checking and typing complex-valued tensors.
    - Concrete subclasses are: CFloatTensor, CHalfTensor, CDoubleTensor.
    - Properties are: is_floating_point=False, is_complex=True, is_signed=True.
    - By default, instantiate this class will create a CFloatTensor.
    """

    _DEFAULT_DTYPE = DTypeEnum.complex64


class ComplexFloatingTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.complex64


class ComplexFloatingTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.complex64


class ComplexFloatingTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.complex64


class ComplexFloatingTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        complex,
        Literal[False],
        Literal[True],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.complex64


class FloatingTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    """Intermediate class for checking and typing floating-point tensors.
    - Concrete subclasses are: FloatTensor, HalfTensor, DoubleTensor.
    - Properties are: is_floating_point=True, is_complex=False, is_signed=True.
    - By default, instantiate this class will create a FloatTensor.
    """

    _DEFAULT_DTYPE = DTypeEnum.float32


class FloatingTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class FloatingTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class FloatingTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class FloatingTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        _float,
        Literal[True],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class SignedIntegerTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    """Intermediate class for checking and typing signed integer data type (integer-like) tensors.
    - Concrete subclasses are: IntTensor, LongTensor, ShortTensor.
    - Properties are: is_floating_point=False, is_complex=False, is_signed=True.
    - By default, instantiate this class will create an IntTensor.
    - BoolTensor is not a subclass of SignedIntegerTensor because it is not signed.
    """

    _DEFAULT_DTYPE = DTypeEnum.int32


class SignedIntegerTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.int32


class SignedIntegerTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.int32


class SignedIntegerTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.int32


class SignedIntegerTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[True],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.int32


class UnsignedIntegerTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    """Intermediate class for checking and typing unsigned integer data type (integer-like) tensors.
    - Concrete subclasses are: BoolTensor, ByteTensor.
    - Properties are: is_floating_point=False, is_complex=False, is_signed=False.
    - By default, instantiate this class will create an ByteTensor.
    - BoolTensor is a subclass of UnsignedIntegerTensor.
    """

    _DEFAULT_DTYPE = DTypeEnum.uint8


class UnsignedIntegerTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.uint8


class UnsignedIntegerTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.uint8


class UnsignedIntegerTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.uint8


class UnsignedIntegerTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        Literal[False],
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.uint8


class IntegralTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
):
    """Intermediate class for checking and typing integer data type (integer-like) tensors.
    - Concrete subclasses are: CharTensor, ShortTensor, IntTensor, LongTensor, BoolTensor, ByteTensor.
    - Properties are: is_floating_point=False, is_complex=False.
    - By default, instantiate this class will create an LongTensor.
    - BoolTensor is a subclass of UnsignedIntegerTensor.
    """

    _DEFAULT_DTYPE = DTypeEnum.long


class IntegralTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.long


class IntegralTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.long


class IntegralTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.long


class IntegralTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        _int,
        Literal[False],
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.long


class RealTensor(
    _TensorNDBase[
        Literal[None],
        _int,
        _int,
        bool,
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        _int,
        _int,
        bool,
        Literal[False],
        _bool,
    ],
):
    """Intermediate class for checking and typing real data type (non complex) tensors.
    - Concrete subclasses are: HalfTensor, FloatTensor, DoubleTensor, CharTensor, ShortTensor, IntTensor, LongTensor, BoolTensor, ByteTensor.
    - Properties are: is_complex=False.
    - By default, instantiate this class will create an FloatTensor.
    """

    _DEFAULT_DTYPE = DTypeEnum.float32


class RealTensor0D(
    _TensorNDBase[
        Literal[None],
        Literal[0],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[0],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class RealTensor1D(
    _TensorNDBase[
        Literal[None],
        Literal[1],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[1],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class RealTensor2D(
    _TensorNDBase[
        Literal[None],
        Literal[2],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[2],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32


class RealTensor3D(
    _TensorNDBase[
        Literal[None],
        Literal[3],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
    metaclass=_TensorNDMeta[
        Literal[None],
        Literal[3],
        _int,
        bool,
        Literal[False],
        _bool,
    ],
):
    _DEFAULT_DTYPE = DTypeEnum.float32



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union

from torch import Tensor

from torchwrench.core.make import DeviceLike, DTypeLike
from torchwrench.core.packaging import _NUMPY_AVAILABLE
from torchwrench.extras.numpy.definitions import np
from torchwrench.extras.numpy.functional import (
    ndarray_to_tensor,
    tensor_to_ndarray,
    to_ndarray,
)
from torchwrench.nn.modules.module import Module


class ToNDArray(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.numpy.to_ndarray`.
    """

    def __init__(
        self,
        *,
        dtype: Union[str, np.dtype, None] = None,
        force: bool = False,
    ) -> None:
        if not _NUMPY_AVAILABLE:
            msg = f"Cannot use {self.__class__.__name__} because numpy dependancy is not installed."
            raise RuntimeError(msg)

        super().__init__()
        self.dtype = dtype
        self.force = force

    def forward(self, x: Union[Tensor, np.ndarray, list]) -> np.ndarray:
        return to_ndarray(x, dtype=self.dtype, force=self.force)


class TensorToNDArray(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.numpy.tensor_to_ndarray`.
    """

    def __init__(
        self,
        *,
        dtype: Union[str, np.dtype, None] = None,
        force: bool = False,
    ) -> None:
        if not _NUMPY_AVAILABLE:
            msg = f"Cannot use {self.__class__.__name__} because numpy dependancy is not installed."
            raise RuntimeError(msg)

        super().__init__()
        self.dtype = dtype
        self.force = force

    def forward(self, x: Tensor) -> np.ndarray:
        return tensor_to_ndarray(x, dtype=self.dtype, force=self.force)


class NDArrayToTensor(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.numpy.ndarray_to_tensor`.
    """

    def __init__(
        self,
        *,
        device: DeviceLike = None,
        dtype: DTypeLike = None,
    ) -> None:
        if not _NUMPY_AVAILABLE:
            msg = f"Cannot use {self.__class__.__name__} because numpy dependancy is not installed."
            raise RuntimeError(msg)

        super().__init__()
        self.device = device
        self.dtype = dtype

    def forward(self, x: np.ndarray) -> Tensor:
        return ndarray_to_tensor(x, dtype=self.dtype, device=self.device)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Final

import torch
from pythonwrench.importlib import is_available_package
from pythonwrench.semver import Version


def _get_extra_version(name: str) -> str:
    try:
        module = __import__(name)
        return str(module.__version__)
    except ImportError:
        return "not_installed"
    except AttributeError:
        return "unknown"


_EXTRAS_PACKAGES = (
    "colorlog",
    "datasets",
    "h5py",
    "numpy",
    "omegaconf",
    "pandas",
    "safetensors",
    "scipy",
    "speechbrain",
    "tensorboard",
    "torchaudio",
    "tqdm",
    "yaml",
)
_EXTRA_AVAILABLE = {name: is_available_package(name) for name in _EXTRAS_PACKAGES}
_EXTRA_VERSION = {name: _get_extra_version(name) for name in _EXTRAS_PACKAGES}


_COLORLOG_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["colorlog"]
_DATASETS_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["datasets"]
_H5PY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["h5py"]
_NUMPY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["numpy"]
_OMEGACONF_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["omegaconf"]
_PANDAS_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["pandas"]
_SAFETENSORS_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["safetensors"]
_SCIPY_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["scipy"]
_SPEECHBRAIN_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["speechbrain"]
_TENSORBOARD_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["tensorboard"]
_TORCHAUDIO_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["torchaudio"]
_TQDM_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["tqdm"]
_YAML_AVAILABLE: Final[bool] = _EXTRA_AVAILABLE["yaml"]


def torch_version_ge_1_13() -> bool:
    version_str = str(torch.__version__)
    version = Version.from_str(version_str)
    return version >= Version("1.13.0")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from dataclasses import asdict
from numbers import Real
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    get_args,
    overload,
)
from typing import Generator as PythonGenerator

import torch
from pythonwrench.collections import all_eq
from pythonwrench.collections import flatten as builtin_flatten
from pythonwrench.collections import prod as builtin_prod
from pythonwrench.functools import (
    function_alias,
    identity,  # noqa: F401
)
from pythonwrench.typing import (
    BuiltinNumber,
    BuiltinScalar,
    SupportsIterLen,
    T_BuiltinScalar,
    is_builtin_scalar,
    is_dataclass_instance,
)
from torch import Tensor, nn
from typing_extensions import Never

from torchwrench.core.make import (
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    as_device,
    as_dtype,
    as_generator,
)
from torchwrench.extras.numpy import np, numpy_view_as_complex, numpy_view_as_real
from torchwrench.nn.functional.cropping import crop_dim
from torchwrench.nn.functional.others import nelement
from torchwrench.nn.functional.padding import PadMode, PadValue, pad_dim
from torchwrench.types import ComplexFloatingTensor, is_builtin_number, is_scalar_like
from torchwrench.types._typing import (
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    ScalarLike,
    T_Tensor,
    T_TensorOrArray,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)
from torchwrench.utils import return_types

T = TypeVar("T")
U = TypeVar("U")

PadCropAlign = Literal["left", "right", "center", "random"]
SqueezeMode = Literal["view_if_possible", "view", "copy", "inplace"]


def repeat_interleave_nd(x: Tensor, repeats: int, dim: int = 0) -> Tensor:
    """Generalized version of torch.repeat_interleave for N >= 1 dimensions.
    The output size will be (..., D*repeats, ...), where D is the size of the dimension of the dim argument.

    Args:
        x: Any tensor of shape (..., D, ...) with at least 1 dim.
        repeats: Number of repeats.
        dim: The dimension to repeat. defaults to 0.

    Examples::
    ----------
        >>> x = torch.as_tensor([[0, 1, 2, 3], [4, 5, 6, 7]])
        >>> repeat_interleave_nd(x, n=2, dim=0)
        tensor([[0, 1, 2, 3],
                [0, 1, 2, 3],
                [4, 5, 6, 7],
                [4, 5, 6, 7]])
    """
    if x.ndim == 0:
        msg = f"Function repeat_interleave_nd does not supports 0-d tensors. (found {x.ndim=} == 0)"
        raise ValueError(msg)

    dim = dim % x.ndim
    x = x.unsqueeze(dim=dim + 1)
    shape = list(x.shape)
    shape[dim + 1] = repeats
    x = x.expand(*shape)
    x = x.flatten(dim, dim + 1)
    return x


def resample_nearest_rates(
    x: Tensor,
    rates: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    """Nearest neigbour resampling using tensor slices.

    Args:
        x: Input tensor.
        rate: The reduction factor of each axis, e.g. a factor of 0.5 will divide the input axes by 2.
        dims: Dimensions to apply resampling. defaults to -1.
        round_fn: Rounding function to compute sub-indices. defaults to torch.floor.
    """
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if isinstance(rates, (int, float)):
        steps = [1.0 / rates] * len(dims)
    else:
        steps = [1.0 / rate for rate in rates]

    return resample_nearest_steps(
        x,
        steps,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_freqs(
    x: Tensor,
    orig_freq: int,
    new_freq: int,
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    """Nearest neigbour resampling using tensor slices.

    Args:
        x: Input tensor.
        orig_freq: Source sampling rate.
        new_freq: Target sampling rate.
        dims: Dimensions to apply resampling. defaults to -1.
        round_fn: Rounding function to compute sub-indices. defaults to torch.floor.
    """
    return resample_nearest_steps(
        x,
        orig_freq / new_freq,
        dims=dims,
        round_fn=round_fn,
    )


def resample_nearest_steps(
    x: Tensor,
    steps: Union[float, Iterable[float]],
    *,
    dims: Union[int, Iterable[int]] = -1,
    round_fn: Callable[[Tensor], Tensor] = torch.floor,
) -> Tensor:
    """Nearest neigbour resampling using tensor slices.

    Args:
        x: Input tensor.
        steps: Floating step for resampling each value.
        dims: Dimensions to apply resampling. defaults to -1.
        round_fn: Rounding function to compute sub-indices. defaults to torch.floor.
    """
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    if isinstance(steps, (int, float)):
        steps = [steps] * len(dims)
    else:
        steps = list(steps)  # type: ignore
        if len(steps) != len(dims):
            raise ValueError(f"Invalid arguments sizes {len(steps)=} != {len(dims)}.")

    slices: List[Union[slice, Tensor]] = [slice(None)] * x.ndim

    for dim, step in zip(dims, steps):
        length = x.shape[dim]
        indexes = torch.arange(0, length, step)
        indexes = round_fn(indexes).long().clamp(min=0, max=length - 1)
        slices[dim] = indexes

    x = x[tuple(slices)]
    return x


def transform_drop(
    transform: Callable[[T], T],
    x: T,
    p: float,
    *,
    generator: GeneratorLike = None,
) -> T:
    """Apply a transform on a tensor with a probability of p.

    Args:
        transform: Transform to apply.
        x: Argument of the transform.
        p: Probability p to apply the transform. Cannot be negative.
            If > 1, it will apply the transform `floor(p)` times and apply a last time with a probability of `p - floor(p)`.
    """
    if p < 0.0:
        raise ValueError(f"Invalid argument {p=} < 0")

    p_floor = math.floor(p)
    for _ in range(p_floor):
        x = transform(x)

    generator = as_generator(generator)
    sampled = torch.rand((), generator=generator)
    if sampled + p_floor < p:
        x = transform(x)

    return x


def pad_and_crop_dim(
    x: Tensor,
    target_length: int,
    *,
    align: PadCropAlign = "left",
    pad_value: PadValue = 0.0,
    dim: int = -1,
    mode: PadMode = "constant",
    generator: GeneratorLike = None,
) -> Tensor:
    """Pad and crop along the specified dimension."""
    x = pad_dim(
        x,
        target_length,
        align=align,
        pad_value=pad_value,
        dim=dim,
        mode=mode,
        generator=generator,
    )
    x = crop_dim(
        x,
        target_length,
        align=align,
        dim=dim,
        generator=generator,
    )
    return x


def shuffled(
    x: T_Tensor,
    dims: Union[int, Iterable[int]] = -1,
    generator: GeneratorLike = None,
) -> T_Tensor:
    """Returns a shuffled version of the input tensor along specific dimension(s)."""
    if isinstance(dims, int):
        dims = [dims]
    else:
        dims = list(dims)

    generator = as_generator(generator)
    slices: List[Union[slice, Tensor]] = [slice(None) for _ in range(x.ndim)]
    for dim in dims:
        indices = torch.randperm(x.shape[dim], generator=generator)
        slices[dim] = indices

    x = x[tuple(slices)]  # type: ignore
    return x


@overload
def flatten(  # type: ignore
    x: Tensor,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> Tensor1D: ...


@overload
def flatten(
    x: Union[np.ndarray, np.generic],
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> np.ndarray: ...


@overload
def flatten(
    x: T_BuiltinScalar,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]: ...


@overload
def flatten(  # type: ignore
    x: Iterable[T_BuiltinScalar],
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> List[T_BuiltinScalar]: ...


def flatten(
    x: Any,
    start_dim: int = 0,
    end_dim: Optional[int] = None,
) -> Union[Tensor1D, np.ndarray, list]:
    if isinstance(x, Tensor):
        end_dim = end_dim if end_dim is not None else x.ndim - 1
        return x.flatten(start_dim, end_dim)  # type: ignore
    elif isinstance(x, np.generic):
        return x.flatten()
    elif isinstance(x, np.ndarray):
        if start_dim == 0 and (end_dim is None or end_dim >= x.ndim - 1):
            return x.flatten()
        else:
            end_dim = end_dim if end_dim is not None else x.ndim - 1
            shape = list(x.shape)
            shape = (
                shape[:start_dim]
                + [builtin_prod(shape[start_dim : end_dim + 1])]
                + shape[end_dim + 1 :]
            )
            return x.reshape(*shape)
    else:
        return builtin_flatten(x, start_dim, end_dim, is_scalar_fn=is_scalar_like)


def squeeze(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int], None] = None,
    mode: SqueezeMode = "view_if_possible",
) -> T_TensorOrArray:
    return _squeeze_impl(x, dim, mode=mode)


def squeeze_(x: Tensor, dim: Union[int, Iterable[int], None] = None) -> Tensor:
    return _squeeze_impl(x, dim, mode="inplace")


def squeeze_copy(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int], None] = None,
) -> T_TensorOrArray:
    return _squeeze_impl(x, dim, mode="copy")


def _squeeze_impl(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int], None],
    mode: SqueezeMode,
) -> T_TensorOrArray:
    if isinstance(x, Tensor):
        return __squeeze_impl_tensor(x, dim, mode)
    elif isinstance(x, np.ndarray):
        return __squeeze_impl_array(x, dim, mode)
    else:
        msg = f"Invalid argument type {type(x)}. (expected Tensor or array)"
        raise TypeError(msg)


def __squeeze_impl_tensor(
    x: Tensor,
    dim: Union[int, Iterable[int], None],
    mode: SqueezeMode,
) -> Tensor:
    if dim is None:
        args = ()
    else:
        args = (dim,)

    if isinstance(dim, int) or dim is None:
        if mode in ("view", "view_if_possible"):
            return torch.squeeze(x, *args)
        elif mode == "copy":
            return torch.squeeze_copy(x, *args)
        elif mode == "inplace":
            return x.squeeze_(*args)
        else:
            msg = f"Invalid argument {mode=}. (expected one of {get_args(SqueezeMode)})"
            raise ValueError(msg)

    else:
        msg = f"Invalid argument type {type(dim)}. (expected int or Iterable[int] or None)"
        raise TypeError(msg)


def __squeeze_impl_array(
    x: np.ndarray,
    dim: Union[int, Iterable[int], None],
    mode: SqueezeMode,
) -> np.ndarray:
    if mode in ("view_if_possible", "copy"):
        if isinstance(dim, Iterable):
            dim = tuple(dim)
        return np.squeeze(x, axis=dim)
    else:
        msg = f"Invalid argument {mode=} with numpy array. (expected one of {('view_if_possible', 'copy')})"
        raise ValueError(msg)


def unsqueeze(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int]],
    mode: SqueezeMode = "view_if_possible",
) -> T_TensorOrArray:
    return _unsqueeze_impl(x, dim, mode=mode)


def unsqueeze_(x: Tensor, dim: Union[int, Iterable[int]]) -> Tensor:
    return _unsqueeze_impl(x, dim, mode="inplace")


def unsqueeze_copy(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int]],
) -> T_TensorOrArray:
    return _unsqueeze_impl(x, dim, mode="copy")


def _unsqueeze_impl(
    x: T_TensorOrArray,
    dim: Union[int, Iterable[int]],
    mode: SqueezeMode,
) -> T_TensorOrArray:
    if isinstance(x, Tensor):
        return __unsqueeze_impl_tensor(x, dim, mode)
    elif isinstance(x, np.ndarray):
        return __unsqueeze_impl_array(x, dim, mode)
    else:
        msg = f"Invalid argument type {type(x)}. (expected Tensor or array)"
        raise TypeError(msg)


def __unsqueeze_impl_tensor(
    x: Tensor,
    dim: Union[int, Iterable[int]],
    mode: SqueezeMode,
) -> Tensor:
    if isinstance(dim, int):
        if mode in ("view", "view_if_possible"):
            return torch.unsqueeze(x, dim)
        elif mode == "copy":
            return torch.unsqueeze_copy(x, dim)
        elif mode == "inplace":
            return x.unsqueeze_(dim)
        else:
            msg = f"Invalid argument {mode=}. (expected one of {get_args(SqueezeMode)})"
            raise ValueError(msg)

    elif isinstance(dim, Iterable):
        for dim_i in dim:
            x = __unsqueeze_impl_tensor(x, dim_i, mode)
        return x

    else:
        msg = f"Invalid argument type {type(dim)}. (expected int or Iterable[int])"
        raise TypeError(msg)


def __unsqueeze_impl_array(
    x: np.ndarray,
    dim: Union[int, Iterable[int]],
    mode: SqueezeMode,
) -> np.ndarray:
    if mode in ("view_if_possible", "copy"):
        if isinstance(dim, Iterable):
            dim = tuple(dim)
        return np.expand_dims(x, axis=dim)
    else:
        msg = f"Invalid argument {mode=} with numpy array. (expected one of {('view_if_possible', 'copy')})"
        raise ValueError(msg)


@overload
def to_item(x: T_BuiltinScalar) -> T_BuiltinScalar: ...


@overload
def to_item(x: Union[Tensor, np.ndarray, SupportsIterLen]) -> BuiltinScalar:  # type: ignore
    ...


def to_item(x: Union[ScalarLike, Tensor, np.ndarray, SupportsIterLen]) -> BuiltinScalar:
    """Convert scalar value to the closest built-in type."""
    if is_builtin_scalar(x, strict=True):
        return x
    elif isinstance(x, (Tensor, np.ndarray, np.generic)) and nelement(x) == 1:
        return x.item()
    elif isinstance(x, SupportsIterLen) and len(x) == 1:
        return to_item(next(iter(x)))
    else:
        msg = f"Invalid argument type {type(x)=}. (expected scalar-like object or an iterable containing only 1 element)"
        raise TypeError(msg)


@overload
def view_as_real(x: Tensor) -> Tensor: ...


@overload
def view_as_real(x: np.ndarray) -> np.ndarray: ...


@overload
def view_as_real(x: complex) -> Tuple[float, float]: ...


def view_as_real(
    x: Union[Tensor, np.ndarray, complex],
) -> Union[Tensor, np.ndarray, Tuple[float, float]]:
    """Convert complex-valued input to floating-point object."""
    if isinstance(x, Tensor):
        return torch.view_as_real(x)
    elif isinstance(x, np.ndarray):
        return numpy_view_as_real(x)
    else:
        return x.real, x.imag


@overload
def view_as_complex(x: Tensor) -> ComplexFloatingTensor: ...


@overload
def view_as_complex(x: np.ndarray) -> np.ndarray: ...


@overload
def view_as_complex(x: Tuple[float, float]) -> complex: ...


def view_as_complex(
    x: Union[Tensor, np.ndarray, Tuple[float, float]],
) -> Union[ComplexFloatingTensor, np.ndarray, complex]:
    """Convert floating-point input to complex-valued object."""
    if isinstance(x, Tensor):
        return torch.view_as_complex(x)  # type: ignore
    elif isinstance(x, np.ndarray):
        return numpy_view_as_complex(x)
    elif (
        isinstance(x, Sequence)
        and len(x) == 2
        and isinstance(x[0], Real)
        and isinstance(x[1], Real)
    ):
        return x[0] + x[1] * 1j
    else:
        msg = f"Invalid argument type {type(x)=}."
        raise TypeError(msg)


@overload
def move_to(
    x: Mapping[T, U],
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Dict[T, U]: ...


@overload
def move_to(
    x: T,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> T: ...


def move_to(
    x: Any,
    predicate: Optional[Callable[[Union[Tensor, nn.Module]], bool]] = None,
    **kwargs,
) -> Any:
    """Move all modules and tensors recursively to a specific dtype or device.

    Works recursively on Iterables, Mappings or dataclasses.
    """
    if "device" in kwargs:
        kwargs["device"] = as_device(kwargs["device"])
    if "dtype" in kwargs:
        kwargs["dtype"] = as_dtype(kwargs["dtype"])

    if isinstance(x, (Tensor, nn.Module)):
        if predicate is None or predicate(x):
            return x.to(**kwargs)
        else:
            return x
    elif is_builtin_scalar(x, strict=True):
        return x
    elif is_dataclass_instance(x):
        return type(x)(**move_to(asdict(x)))
    elif isinstance(x, Mapping):
        return {k: move_to(v, predicate=predicate, **kwargs) for k, v in x.items()}
    elif isinstance(x, Iterable):
        generator = (move_to(xi, predicate=predicate, **kwargs) for xi in x)
        if isinstance(x, PythonGenerator):
            return generator
        elif isinstance(x, (tuple, set, frozenset)):
            return type(x)(generator)
        else:
            return list(generator)
    else:
        return x


@function_alias(move_to)
def move_to_rec(*args, **kwargs): ...


@function_alias
def recursive_to(*args, **kwargs): ...


# ----------
# as_tensor
# ----------


# Empty lists
@overload
def as_tensor(  # type: ignore
    data: Sequence[Never],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Never]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[Never]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> Tensor3D: ...


# bool
@overload
def as_tensor(  # type: ignore
    data: bool,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor0D: ...


@overload
def as_tensor(  # type: ignore
    data: Sequence[bool],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[bool]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[bool]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> BoolTensor3D: ...


@overload
def as_tensor(  # type: ignore
    data: BuiltinNumber,
    dtype: Literal["bool"],
    device: DeviceLike = None,
) -> BoolTensor0D: ...


@overload
def as_tensor(  # type: ignore
    data: Sequence[BuiltinNumber],
    dtype: Literal["bool"],
    device: DeviceLike = None,
) -> BoolTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: Literal["bool"],
    device: DeviceLike = None,
) -> BoolTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: Literal["bool"],
    device: DeviceLike = None,
) -> BoolTensor3D: ...


# int
@overload
def as_tensor(
    data: int,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor0D: ...


@overload
def as_tensor(
    data: Sequence[int],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[int]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[int]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> LongTensor3D: ...


@overload
def as_tensor(
    data: BuiltinNumber,
    dtype: Literal["int64", "long"],
    device: DeviceLike = None,
) -> LongTensor0D: ...


@overload
def as_tensor(
    data: Sequence[BuiltinNumber],
    dtype: Literal["int64", "long"],
    device: DeviceLike = None,
) -> LongTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: Literal["int64", "long"],
    device: DeviceLike = None,
) -> LongTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: Literal["int64", "long"],
    device: DeviceLike = None,
) -> LongTensor3D: ...


# float
@overload
def as_tensor(
    data: float,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor0D: ...


@overload
def as_tensor(
    data: Sequence[float],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[float]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[float]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> FloatTensor3D: ...


@overload
def as_tensor(
    data: BuiltinNumber,
    dtype: Literal["float32", "float"],
    device: DeviceLike = None,
) -> FloatTensor0D: ...


@overload
def as_tensor(
    data: Sequence[BuiltinNumber],
    dtype: Literal["float32", "float"],
    device: DeviceLike = None,
) -> FloatTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: Literal["float32", "float"],
    device: DeviceLike = None,
) -> FloatTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: Literal["float32", "float"],
    device: DeviceLike = None,
) -> FloatTensor3D: ...


# complex
@overload
def as_tensor(
    data: complex,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor0D: ...


@overload
def as_tensor(
    data: Sequence[complex],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[complex]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[complex]]],
    dtype: Literal[None] = None,
    device: DeviceLike = None,
) -> CFloatTensor3D: ...


@overload
def as_tensor(
    data: BuiltinNumber,
    dtype: Literal["complex64", "cfloat"],
    device: DeviceLike = None,
) -> CFloatTensor0D: ...


@overload
def as_tensor(
    data: Sequence[BuiltinNumber],
    dtype: Literal["complex64", "cfloat"],
    device: DeviceLike = None,
) -> CFloatTensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: Literal["complex64", "cfloat"],
    device: DeviceLike = None,
) -> CFloatTensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: Literal["complex64", "cfloat"],
    device: DeviceLike = None,
) -> CFloatTensor3D: ...


# BuiltinNumber
@overload
def as_tensor(
    data: BuiltinNumber,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor0D: ...


@overload
def as_tensor(
    data: Sequence[BuiltinNumber],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor1D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[BuiltinNumber]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor2D: ...


@overload
def as_tensor(
    data: Sequence[Sequence[Sequence[BuiltinNumber]]],
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> Tensor3D: ...


@overload
def as_tensor(
    data: Any,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> torch.Tensor: ...


def as_tensor(
    data: Any,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
) -> torch.Tensor:
    """Convert arbitrary data to tensor.

    Unlike `torch.as_tensor`, it works recursively and stack sequences like List[Tensor]. It also accept python generator objects.

    Args:
        data: Data to convert to tensor. Can be Tensor, np.ndarray, list, tuple or any number-like object.
        dtype: Target torch dtype. defaults to None.
        device: Target torch device. defaults to None.

    Returns:
        PyTorch tensor created from data.
    """
    if isinstance(data, (Tensor, np.ndarray, np.number)) or is_builtin_number(data):
        dtype = as_dtype(dtype)
        device = as_device(device)
        return torch.as_tensor(data, dtype=dtype, device=device)

    elif isinstance(data, (list, tuple, PythonGenerator)):
        dtype = as_dtype(dtype)
        device = as_device(device)

        tensors: List[Tensor] = [
            as_tensor(data_i, dtype=dtype, device=device) for data_i in data
        ]
        if len(tensors) == 0:
            return torch.as_tensor(tensors, dtype=dtype, device=device)

        shapes = [tensor.shape for tensor in tensors]
        if not all_eq(shapes):
            uniq_shapes = tuple(set(shapes))
            msg = f"Cannot convert to tensor a list of elements with heterogeneous shapes. (found different shapes: {uniq_shapes})"
            raise ValueError(msg)

        return torch.stack(tensors)

    else:
        EXPECTED = (
            Tensor,
            np.ndarray,
            np.number,
            BuiltinNumber,
            list,
            tuple,
            PythonGenerator,
        )
        msg = f"Invalid argument type '{type(data)}'. (expected one of {EXPECTED})"
        raise TypeError(msg)


@overload
def topk(
    x: Tensor,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
    *,
    return_values: Literal[True] = True,
    return_indices: Literal[True] = True,
) -> return_types.topk: ...


@overload
def topk(
    x: T_Tensor,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
    *,
    return_values: Literal[True] = True,
    return_indices: Literal[False],
) -> T_Tensor: ...


@overload
def topk(
    x: Tensor,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
    *,
    return_values: Literal[False],
    return_indices: Literal[True] = True,
) -> LongTensor: ...


@overload
def topk(
    x: T_Tensor,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
    *,
    return_values: bool = True,
    return_indices: bool = True,
) -> Union[T_Tensor, LongTensor, return_types.topk]: ...


def topk(
    x: T_Tensor,
    k: int,
    dim: int = -1,
    largest: bool = True,
    sorted: bool = True,
    *,
    return_values: bool = True,
    return_indices: bool = True,
) -> Union[T_Tensor, LongTensor, return_types.topk]:
    result = x.topk(
        k=k,
        dim=dim,
        largest=largest,
        sorted=sorted,
    )
    if return_values and return_indices:
        return result  # type: ignore
    elif return_values:
        return result.values  # type: ignore
    elif return_indices:
        return result.indices  # type: ignore
    else:
        msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
        raise ValueError(msg)


@overload
def top_p(
    x: Tensor,
    p: float,
    dim: int = -1,
    largest: bool = True,
    *,
    return_values: Literal[True] = True,
    return_indices: Literal[True] = True,
) -> return_types.top_p: ...


@overload
def top_p(
    x: T_Tensor,
    p: float,
    dim: int = -1,
    largest: bool = True,
    *,
    return_values: Literal[True] = True,
    return_indices: Literal[False],
) -> T_Tensor: ...


@overload
def top_p(
    x: Tensor,
    p: float,
    dim: int = -1,
    largest: bool = True,
    *,
    return_values: Literal[False],
    return_indices: Literal[True] = True,
) -> LongTensor: ...


@overload
def top_p(
    x: T_Tensor,
    p: float,
    dim: int = -1,
    largest: bool = True,
    *,
    return_values: bool = True,
    return_indices: bool = True,
) -> Union[T_Tensor, LongTensor, return_types.top_p]: ...


def top_p(
    x: T_Tensor,
    p: float,
    dim: int = -1,
    largest: bool = True,
    *,
    return_values: bool = True,
    return_indices: bool = True,
) -> Union[T_Tensor, LongTensor, return_types.top_p]:
    values, indices = x.sort(dim=dim, descending=largest)
    cumulated = torch.cumsum(values, dim=dim)
    idx = (cumulated >= p).long().argmax()
    result = return_types.top_p([values[:idx], indices[:idx]])

    if return_values and return_indices:
        return result  # type: ignore
    elif return_values:
        return result.values  # type: ignore
    elif return_indices:
        return result.indices  # type: ignore
    else:
        msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
        raise ValueError(msg)


@function_alias(as_tensor)
def to_tensor(*args, **kwargs): ...


@function_alias(topk)
def top_k(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from typing import Iterator
from unittest import TestCase

from torchwrench.utils.data.dataset.wrapper import (
    IterableSubset,
    IterableTransformWrapper,
    TransformWrapper,
)


class TestDataset(TestCase):
    def test_advanced_example_1(self) -> None:
        class Dummy:
            def __iter__(self) -> Iterator[int]:
                yield from range(len(self))

            def __len__(self) -> int:
                return 100

        ds = Dummy()
        subset = IterableSubset(ds, [2, 4, 6])

        assert list(subset) == [2, 4, 6]

    def test_wrapper_example_1(self) -> None:
        wrapped = IterableTransformWrapper(list(range(10)), lambda x: x * 2)
        assert list(wrapped) == list(range(0, 20, 2))

    def test_transform_wrapper(self) -> None:
        wrapped = TransformWrapper(list(range(10)), lambda x: x * 2)
        assert [wrapped[i] for i in range(len(wrapped))] == list(range(0, 20, 2))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# for backward compatibility
from torchwrench.nn.functional.new import (  # noqa: F401
    arange,
    empty,
    full,
    ones,
    rand,
    randint,
    randperm,
    zeros,
)
from torchwrench.nn.functional.transform import (  # noqa: F401
    as_tensor,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchwrench.core.packaging import _SPEECHBRAIN_AVAILABLE

if not _SPEECHBRAIN_AVAILABLE:
    from torchwrench.extras.speechbrain._speechbrain_fallback import DynamicItemDataset

else:
    from speechbrain.dataio.dataset import DynamicItemDataset


__all__ = [
    "_SPEECHBRAIN_AVAILABLE",
    "DynamicItemDataset",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from yaml import SafeLoader  # noqa: F401

from torchwrench.extras.yaml import (  # noqa: F401
    IgnoreTagLoader,
    SplitTagLoader,
    dump_yaml,
    dumps_yaml,
    load_yaml,
    loads_yaml,
    read_yaml,
    save_yaml,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from email.message import Message
from pathlib import Path
from typing import Optional, Union
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

# for backward compatibility
from pythonwrench.hashlib import hash_file  # noqa: F401
from pythonwrench.os import safe_rmdir  # noqa: F401
from pythonwrench.warnings import warn_once
from torch.hub import download_url_to_file

pylog = logging.getLogger(__name__)


def download_file(
    url: str,
    dst: Union[str, Path, None] = ".",
    *,
    hash_prefix: Optional[str] = None,
    make_parents: bool = False,
    make_intermediate: Optional[bool] = None,
    verbose: int = 0,
) -> Path:
    """Download file to target filepath or directory.

    Args:
        url: Target URL.
        dst: Target filepath or directory. None means current working directory. defaults to ".".
        hash_prefix: Optional hash prefix present in destination filename. defaults to None.
        make_parents: If True, make intermediate directories to destination. defaults to False.
        make_intermediate: Deprecated: alias for 'make_parents'. If not None, overwrite any value of 'make_parents'. defaults to None.
        verbose: Verbose level. defaults to 0.

    Returns:
        Path to the downloaded file.
    """
    if make_intermediate is not None:
        msg = f"Deprecated argument {make_intermediate=}. Use make_parents={make_intermediate} instead."
        warn_once(msg)
        make_parents = make_intermediate
    del make_intermediate

    if dst is None:
        dst = "."
    dst = Path(dst)

    if dst.is_dir():
        fname = _get_filename_from_url(url)
        fpath = dst.joinpath(fname)
    elif dst.exists() and not dst.is_file():
        msg = f"Destination '{dst}' exists but is not a file or directory."
        raise FileExistsError(msg)
    else:  # is_file or not exists
        fpath = dst
    del dst

    if make_parents:
        dpath = fpath.parent
        dpath.mkdir(parents=True, exist_ok=True)

    try:
        download_url_to_file(
            url,
            str(fpath),
            hash_prefix=hash_prefix,
            progress=verbose > 0,
        )

    except HTTPError as err:
        msg = f"Cannot download from {url=}. (with {fpath=}, {hash_prefix=}, {make_parents=})"
        pylog.error(msg)
        raise err

    return fpath


def _get_filename_from_url(url: str, timeout_s: Optional[float] = None) -> str:
    try:
        response = urlopen(url, timeout=timeout_s)
        header = response.headers.get("Content-Disposition", "")
        message = Message()
        message["content-type"] = header
        filename = message.get_param("filename", None)
    except (URLError, ValueError):
        msg = (
            f"Cannot get target filename from {url=}. Try to detect it from URL string."
        )
        pylog.warning(msg)
        filename = None

    if filename is None:
        filename = url.split("/")[-1].split("?")[0]
    return str(filename)
